# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mock
import os
import unittest
import shutil
import tempfile

import yaml

import hooks


PATCHES = [
    'charmhelpers.core.hookenv.close_port',
    'charmhelpers.core.hookenv.config',
    'charmhelpers.core.hookenv.log',
    'charmhelpers.core.hookenv.open_port',
    'charmhelpers.core.hookenv.relation_get',
    'charmhelpers.core.hookenv.relation_id',
    'charmhelpers.core.hookenv.relation_set',
    'charmhelpers.core.hookenv.service_name',
    'charmhelpers.core.hookenv.unit_get',
    'charmhelpers.core.host.mkdir',
    'charmhelpers.fetch.apt_install',
    'charmhelpers.fetch.configure_sources',
    'charmhelpers.payload.execd.execd_preinstall',
]


class JujuTestCase(unittest.TestCase):
    def setUp(self):
        """Setup a testing hook environment.

        Patches the required charmhelper and intercom objects and
        initializes few mocks for tests.
        """
        super(JujuTestCase, self).setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)

        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        self.relation_id.return_value = '1'
        self.service_name.return_value = 'rabbit-worker'
        self.unit_get.return_value = '127.0.0.1'
        self._set_default_config()
        os.mkdir(hooks._service_dir())

    def _set_default_config(self):
        defaults = {}
        config = os.path.join(os.path.dirname(__file__), '../config.yaml')
        if os.path.exists(config):
            with open(config) as f:
                config = yaml.safe_load(f)
                for name, option in config['options'].iteritems():
                    defaults[name] = option.get('default')
        defaults['install_root'] = self.tmpdir
        self.config.return_value = defaults


class TestInstallHook(JujuTestCase):
    def setUp(self):
        super(TestInstallHook, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)
        os.mkdir(os.path.join(self.tmpdir, 'code'))

    def test_payload_local(self):
        self.assertIsNone(hooks._payload_local('http://no_such_sha1.tgz'))
        name = 'r123_sha1hash'
        os.mkdir(hooks._code_dir(name))
        self.assertEqual(name, hooks._payload_local('http://sha1hash.tgz'))

    def test_payload_not_configured(self):
        local = ['r1_aa', 'r2_bb', 'r3_cc', 'r4_dd']
        configured = ['http://aa.tgz', 'http://cc.tgz', 'http://dd.tgz']
        os.mkdir(hooks._code_dir())  # create a directory for "current"
        for x in local:
            os.mkdir(hooks._code_dir(x))
        not_configured = hooks._payloads_not_configured(configured)
        self.assertEquals(not_configured, set(['r2_bb']))

    @mock.patch('hooks._payload_extract')
    def test_handle_code_noop(self, payload_extract):
        local = ['r1_aa', 'r2_bb']
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = '''
            http://aa.tgz http://bb.tgz'''
        for x in local:
            os.mkdir(hooks._code_dir(x))
        hooks.handle_code()
        self.assertEqual(0, payload_extract.call_count)
        self.assertEqual('r2_bb', os.readlink(hooks._code_dir()))

    @mock.patch('hooks._payload_extract')
    def test_handle_code(self, payload_extract):
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = 'http://bb.tgz'

        payload_extract.side_effect = lambda x: os.mkdir(
            hooks._code_dir('r2_bb'))
        self.assertEqual(set([]), hooks.handle_code())
        self.assertEqual(1, payload_extract.call_count)
        self.assertEqual('r2_bb', os.readlink(hooks._code_dir()))


class TestAMQPHooks(JujuTestCase):
    def setUp(self):
        super(TestAMQPHooks, self).setUp()
        self.cfg_file = os.path.join(
            hooks._service_dir(), 'amqp_config.py')

    def test_joined(self):
        hooks.hooks.execute(['hooks/amqp-relation-joined'])
        self.assertEqual(1, self.relation_set.call_count)

    @mock.patch('subprocess.call')
    def test_broken(self, subprocess):
        with open(self.cfg_file, 'w') as f:
            f.write('')
            f.flush()
        hooks.hooks.execute(['hooks/amqp-relation-broken'])
        self.assertFalse(os.path.exists(self.cfg_file))

    @mock.patch('subprocess.check_call')
    def test_changed_ready(self, check_call):
        self.relation_get.return_value = {
            'password': 'foo',
            'private-address': '1.1.1.1',
        }
        hooks.hooks.execute(['hooks/amqp-relation-changed'])
        self.assertTrue(os.path.exists(self.cfg_file))
        check_call.assert_called_once_with(['/sbin/restart', 'rabbit-worker'])

    @mock.patch('subprocess.check_call')
    def test_changed_notready(self, check_call):
        self.relation_get.return_value = {
            'private-address': '1.1.1.1',
        }
        hooks.hooks.execute(['hooks/amqp-relation-changed'])
        self.assertFalse(os.path.exists(self.cfg_file))
        self.assertEqual(0, check_call.call_count)


class TestCronConfig(JujuTestCase):
    @mock.patch('os.symlink')
    @mock.patch('os.unlink')
    def test_cron_unconfigured(self, unlink, symlink):
        hooks._update_cron_config()
        self.assertEqual(0, symlink.call_count)
        self.assertEqual(0, unlink.call_count)

    @mock.patch('os.symlink')
    @mock.patch('os.unlink')
    def test_cron_good(self, unlink, symlink):
        self.config.return_value['cron_period'] = 'daily'
        self.config.return_value['cron_cmd'] = 'foo.sh'
        hooks._update_cron_config()
        self.assertEqual(0, unlink.call_count)
        symlink.assert_called_once_with(
            os.path.join(self.tmpdir, 'rabbit-worker/code/current/foo.sh'),
            '/etc/cron.daily/rabbit-worker')

    def test_cron_bad(self):
        self.config.return_value['cron_period'] = 'invalid'
        with self.assertRaises(ValueError):
            hooks._update_cron_config()


class TestJSONStatusHooks(JujuTestCase):
    def test_joined(self):
        hooks.hooks.execute(['hooks/json_status-relation-joined'])
        self.relation_set.assert_called_once_with(
            None, {'status-url': 'http://127.0.0.1:8080/json_status'})
