#!/usr/bin/env python

import base64
import json
import os
import re
import shutil
import socket
import subprocess
import sys
import tempfile
import time

from charmhelpers.contrib.charmsupport.nrpe import NRPE

import charmhelpers.core.hookenv
import charmhelpers.core.host
import charmhelpers.fetch
import charmhelpers.payload.execd

hooks = charmhelpers.core.hookenv.Hooks()


def juju_info(msg):
    charmhelpers.core.hookenv.log(msg, 'INFO')


def _service_dir(config):
    unit = charmhelpers.core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(config['install_root'], unit)


def _code_dir(config, revno='current'):
    return os.path.join(_service_dir(config), 'code/%s' % revno)


def _get_revno(src):
    revno = os.path.basename(src).split('.tgz')[0]
    revfile = os.path.join(src, '.uploaded_version_info')
    if os.path.exists(revfile):
        # revline will be: branch@617 - 2014-06-25 11:50:44 -0500
        revline = open(revfile).read()
        match = re.match(r'.*@(\d+)', revline)
        if match:
            revno = 'r' + match.group(1) + '_' + revno
    return revno


def _install_tarball(config):
    current_url = _code_dir(config, '.current-tarball-url')
    if os.path.exists(current_url):
        url = open(current_url).read()
        if url == config['tarball']:
            return

    juju_info('grabbing service from tarball...')
    src = charmhelpers.fetch.install_remote(config['tarball'])

    revno = _get_revno(src)

    charmhelpers.core.host.mkdir(_code_dir(config, '..'))

    # charmhelpers puts the payload in an odd place, move it to our code_dir
    dst = _code_dir(config, revno)
    shutil.move(src, dst)

    current = _code_dir(config)
    if os.path.exists(current):
        os.unlink(current)
    juju_info('linking %s as current' % dst)
    os.symlink(dst, current)

    with open(current_url, 'w') as f:
        f.write(config['tarball'])
    return True


@hooks.hook()
def install(upgrade=False):
    # Support Canonistack/Prodstack basenode subordinate charm interface.
    charmhelpers.payload.execd.execd_preinstall()
    config = charmhelpers.core.hookenv.config()

    pkgs = [x for x in config['packages'].split(' ') if x]
    pkgs.append('bzr')

    framework = config['framework']
    if framework == 'restish':
        pkgs.append('python-restish')
    elif framework == 'pyramid':
        pkgs.append('python3-pyramid')
    elif framework == 'django':
        pkgs.append('python-django')
    else:
        juju_info('Invalid framework specified: {}'.format(framework))
        sys.exit(1)

    juju_info('installing apt packages...')
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(pkgs, fatal=True)


@hooks.hook()
def upgrade_charm():
    install(True)


def _wsgi_reload():
    # use upstart if possible or in the context of a non-wsgi relation
    if os.path.exists('/etc/init/gunicorn.conf'):
        charmhelpers.core.host.service_restart('gunicorn')
        return

    # Trigger WSGI reloading
    for relid in charmhelpers.core.hookenv.relation_ids():
        charmhelpers.core.hookenv.relation_set(
            relid, {'wsgi_timestamp': time.time()})


def _update_cron_config():
    periods = ('hourly', 'daily', 'weekly')
    service = charmhelpers.core.hookenv.service_name()
    config = charmhelpers.core.hookenv.config()

    if config['cron_period'] not in periods:
        raise ValueError('Invalid cron_peroid: %s' % config['cron_period'])

    for period in periods:
        p = '/etc/cron.%s/%s' % (period, service)
        if os.path.exists(p) or os.path.lexists(p):
            juju_info('removing old cron config: %s' % p)
            os.unlink(p)

    cmd = config['cron_cmd']
    if cmd:
        cmd = os.path.abspath(os.path.join(_code_dir(config), cmd))
        juju_info('installing cron job for: %s' % cmd)
        os.symlink(cmd, '/etc/cron.%s/%s' % (config['cron_period'], service))


def _set_allowed_hosts():
    config = charmhelpers.core.hookenv.config()
    ips = [
        socket.gethostname(),
        socket.getfqdn(),
        charmhelpers.core.hookenv.unit_get('public-address'),
        charmhelpers.core.hookenv.unit_private_ip(),
        'localhost',
    ]
    host = config.get('hostname')
    if host:
        ips.append(host)

    root_dir = _service_dir(config)
    with open(os.path.join(root_dir, 'allowed_hosts.json'), 'w') as f:
        json.dump(ips, f)


@hooks.hook()
def config_changed():
    config = charmhelpers.core.hookenv.config()

    _update_cron_config()
    if _install_tarball(config):
        _db_migrate()
    _set_allowed_hosts()

    unit_config = config.get('unit-config')
    if unit_config is not None:
        root_dir = _service_dir(config)
        with open(os.path.join(root_dir, 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))

    update_nrpe_config()
    _wsgi_reload()


@hooks.hook('website-relation-changed')
@hooks.hook()
def website_relation_joined():
    config = charmhelpers.core.hookenv.config()
    host = subprocess.check_output(['unit-get', 'private-address']).strip()
    charmhelpers.core.hookenv.relation_set(
        None, {'port': config["port"], 'hostname': host})


@hooks.hook('json_status-relation-joined')
def json_status_relation_joined():
    config = charmhelpers.core.hookenv.config()
    path = config.get('json_status_path')
    if path:
        port = config['port']
        host = charmhelpers.core.hookenv.unit_get('public-address')
        if not host:
            juju_info('no public address found, using private')
            host = charmhelpers.core.hookenv.unit_private_ip()
        url = 'http://%s:%s/%s' % (host, port, path)
        charmhelpers.core.hookenv.relation_set(None, {'status-url': url})


@hooks.hook('wsgi-relation-changed')
@hooks.hook()
def wsgi_relation_joined():
    config = charmhelpers.core.hookenv.config()
    cdir = _code_dir(config)
    ldir = os.path.join(_service_dir(config), 'logs')

    if not os.path.exists(ldir):
        charmhelpers.core.host.mkdir(ldir, 'www-data', 'www-data', 0755)

    full_path = []
    path = config.get('python_path')
    if path:
        full_path = []
        for p in path.split(':'):
            if p[0] != '/':
                p = os.path.join(cdir, p)
            full_path.append(p)

    charmhelpers.core.hookenv.relation_set(relation_settings={
        'working_dir': cdir,
        'wsgi_log_file': os.path.join(ldir, 'gunicorn.log'),
        'wsgi_access_logfile': os.path.join(ldir, 'access.log'),
        'port': config['port'],
        'python_path': ':'.join(full_path),
    })

    charmhelpers.core.hookenv.open_port(config['port'])


@hooks.hook()
def wsgi_relation_broken():
    port = charmhelpers.core.hookenv.config()['port']
    charmhelpers.core.hookenv.close_port(port)


@hooks.hook('nrpe-external-master-relation-changed')
@hooks.hook()
def nrpe_external_master_relation_joined():
    update_nrpe_config()


@hooks.hook('local-monitors-relation-changed')
@hooks.hook()
def local_monitors_relation_joined():
    update_nrpe_config()


def update_nrpe_config():
    nrpe_compat = NRPE()
    conf = nrpe_compat.config
    check_http_params = conf.get('nagios_check_http_params')
    if check_http_params:
        nrpe_compat.add_check(
            shortname='restish',
            description='Check Restish Server',
            check_cmd='check_http %s' % check_http_params
        )
    nrpe_compat.write()


def _db_migrate():
    config = charmhelpers.core.hookenv.config()
    if not os.path.exists(_pgsql_file(config)):
        # the pgsql relation doesn't exist
        return
    cmd = config.get('db_migration_cmd')
    if cmd:
        subprocess.check_call([cmd], shell=True, cwd=_code_dir(config))


def _pgsql_file(config):
    return os.path.join(_service_dir(config), 'pgsql.json')


@hooks.hook('pgsql-relation-joined')
@hooks.hook('pgsql-relation-changed')
def pgsql_relation_joined_changed():
    config = charmhelpers.core.hookenv.config()
    packages = ["python-psycopg2", "postgresql-client"]
    charmhelpers.fetch.apt_install(packages, fatal=True)

    relation = charmhelpers.core.hookenv.relation_get()
    with open(_pgsql_file(config), 'w') as f:
        json.dump(relation, f)

    # some relation calls will just have "private-address". We need to ensure
    # we've gotten all the db credentials before trying to migrate
    if relation.get('user') and relation.get('host'):
        _db_migrate()
    _wsgi_reload()


@hooks.hook()
def pgsql_relation_broken():
    config = charmhelpers.core.hookenv.config()
    settings = _pgsql_file(config)
    if os.path.exists(settings):
        os.unlink(settings)

    _wsgi_reload()


def _create_amqp_config(config):
    path = os.path.join(_service_dir(config), 'amqp_config.py')
    relation = charmhelpers.core.hookenv.relation_get()
    if 'password' not in relation:
        return path  # not yet ready

    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write('# DO NOT EDIT. Generated by rabbitmq-worker charm hook\n')
        f.write('AMQP_USER = "%s"\n' % config['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n' % config['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % relation['private-address'])
        f.write('AMQP_PASSWORD = "%s"\n' % relation['password'])
    # Now that the file is complete (and closed) rename it to its final
    # destination. This avoids races where another process try to import that
    # file as a python module (creating a .pyc) and ends up having close enough
    # time stamps to fool python (which won't reload the completed file).
    os.rename(f.name, path)
    os.chmod(path, 0644)
    return path


@hooks.hook()
def amqp_relation_joined():
    config = charmhelpers.core.hookenv.config()
    charmhelpers.core.hookenv.relation_set(
        None, {'username': config['amqp-user'], 'vhost': config['amqp-vhost']})


@hooks.hook()
def amqp_relation_changed():
    config = charmhelpers.core.hookenv.config()
    if not os.path.exists(_create_amqp_config(config)):
        return
    _wsgi_reload()


@hooks.hook()
def amqp_relation_departed():
    config = charmhelpers.core.hookenv.config()
    os.unlink(os.path.join(_service_dir(config), 'amqp_config.py'))
    _wsgi_reload()


if __name__ == "__main__":
    hooks.execute(sys.argv)
