# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import yaml

from ucitests import fixtures

from ci_utils.testing import (
    features,
    fixtures as ci_fixtures,
)


class TestSwift(unittest.TestCase):

    def test_no_env_vars_no_swift(self):
        fixtures.isolate_from_env(self, dict(OS_AUTH_URL=None,
                                             OS_USERNAME=None,
                                             OS_PASSWORD=None,
                                             OS_TENANT_NAME=None,
                                             OS_REGION_NAME=None))
        swift = features.Swift()
        self.assertFalse(swift.available())


class TestNovaCompute(unittest.TestCase):

    def test_no_env_vars_no_nova(self):
        fixtures.isolate_from_env(self, dict(GLANCE_OS_USERNAME=None,
                                             GLANCE_OS_PASSWORD=None,
                                             GLANCE_OS_TENANT_NAME=None,
                                             GLANCE_OS_AUTH_URL=None,
                                             GLANCE_OS_REGION_NAME=None))
        nova = features.NovaCompute()
        self.assertFalse(nova.available())


class TestJujuHome(unittest.TestCase):

    def setUp(self):
        super(TestJujuHome, self).setUp()
        ci_fixtures.isolate_from_juju_env(self)
        self.bsj = features.BootstrappedJuju()

    def test_no_juju_home(self):
        del os.environ['JUJU_HOME']
        self.assertEqual(os.path.join(self.uniq_dir, '.juju'),
                         self.bsj.get_juju_home())

    def test_juju_home_set(self):
        os.environ['JUJU_HOME'] = 'foo'
        self.assertEqual('foo', self.bsj.get_juju_home())


class TestJujuGetEnvName(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnvName, self).setUp()
        ci_fixtures.isolate_from_juju_env(self)
        self.bsj = features.BootstrappedJuju()

    def test_no_env_name(self):
        self.assertEqual(None, self.bsj.get_env_name())

    def test_env_name_from_os_env(self):
        os.environ['JUJU_ENV'] = 'foo'
        self.assertEqual('foo', self.bsj.get_env_name())

    def test_no_env_name_in_env_yaml(self):
        fixtures.build_tree('''file: environments.yaml
''')
        self.assertEqual(None, self.bsj.get_env_name())

    def test_env_name_from_current_file(self):
        fixtures.build_tree('''file: current-environment
foo
''')
        self.assertEqual('foo', self.bsj.get_env_name())

    def test_env_name_from_env_yaml(self):
        fixtures.build_tree('''file: environments.yaml
default: foo
''')
        self.assertEqual('foo', self.bsj.get_env_name())

    def test_os_env_override_current_file(self):
        fixtures.build_tree('''file: environments.yaml
default: foo
''')
        os.environ['JUJU_ENV'] = 'bar'
        self.assertEqual('bar', self.bsj.get_env_name())

    def test_current_override_yaml_file(self):
        fixtures.build_tree('''file: environments.yaml
default: foo
file: current-environment
bar
''')
        self.assertEqual('bar', self.bsj.get_env_name())

    def test_no_control_bucket(self):
        fixtures.build_tree('''file: environments.yaml
default: foo
environments:
    foo:
''')
        self.assertEqual(None, self.bsj.get_control_bucket_id())


class TestJujuGetEnvsFromYaml(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnvsFromYaml, self).setUp()
        ci_fixtures.isolate_from_juju_env(self)
        self.bsj = features.BootstrappedJuju()

    def test_no_environments_file(self):
        self.assertEqual(None, self.bsj.get_environments_from_yaml())

    def test_broken_environments_file(self):
        # a juju file that defines no environment
        fixtures.build_tree('''file: environments.yaml
default: foo
''')
        self.assertFalse(None, self.bsj.available())

    def test_inconsistent_environments_file(self):
        # a juju file that defines a default environment that doesn't exist
        fixtures.build_tree('''file: environments.yaml
default: foo
environments:
    bar:
        type: local
''')
        self.assertFalse(None, self.bsj.available())


class TestJujuGetEnv(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnv, self).setUp()
        ci_fixtures.isolate_from_juju_env(self)
        self.bsj = features.BootstrappedJuju()

    def test_no_env(self):
        self.assertEqual(None, self.bsj.get_env('foo'))

    def test_broken_env_file(self):
        fixtures.build_tree('''file: environments.yaml
''')
        self.assertEqual(None, self.bsj.get_env('foo'))

    def test_unknown_env(self):
        fixtures.build_tree('''file: environments.yaml
environments:
    bar:
        type: local
''')
        self.assertEqual(None, self.bsj.get_env('foo'))

    def test_get_one_env(self):
        fixtures.build_tree('''file: environments.yaml
environments:
    bar: {}
''')
        self.assertEqual({}, self.bsj.get_env('bar'))


class TestProviderType(unittest.TestCase):

    def setUp(self):
        super(TestProviderType, self).setUp()
        ci_fixtures.isolate_from_juju_env(self)
        # We don't want to test the specific providers, we only care about
        # finding their type. So we patch the type specific checks to be always
        # successful and test available() from a give environments.yaml file.

        def probe_true(feature):
            return True

        fixtures.patch(self, features.BootstrappedJuju,
                       '_probe_local', probe_true)
        fixtures.patch(self, features.BootstrappedJuju,
                       '_probe_openstack', probe_true)
        self.bsj = features.BootstrappedJuju()

    def test_no_env(self):
        self.assertFalse(self.bsj.available())

    def test_local_provider(self):
        fixtures.build_tree('''file: environments.yaml
default: bar
environments: {bar: {type: local}}
''')
        self.assertTrue(self.bsj.available())

    def test_openstack_provider(self):
        fixtures.build_tree('''file: environments.yaml
default: bar
environments: {bar: {type: openstack}}
''')
        self.assertTrue(self.bsj.available())

    def test_unknown_provider(self):
        fixtures.build_tree('''file: environments.yaml
default: bar
environments: {bar: {type: vaporware}}
''')
        self.assertFalse(self.bsj.available())


class TestJujuNova(unittest.TestCase):

    def setUp(self):
        super(TestJujuNova, self).setUp()
        self.bsj = features.BootstrappedJuju()

    def test_get_nova_id_well_formed(self):
        provider_state = yaml.safe_load('''state-instances:
- a8c77d73-55b5-4b21-8c43-bf5f60ad5ecb
characteristics:
- arch: amd64
  mem: 1024
  rootdisk: 10240
  cpucores: 1
''')
        self.assertEqual('a8c77d73-55b5-4b21-8c43-bf5f60ad5ecb',
                         self.bsj.get_nova_id(provider_state))

    def test_get_nova_id_malformed(self):
        self.assertEqual(None, self.bsj.get_nova_id(None))
        self.assertEqual(None,
                         self.bsj.get_nova_id({'state-instances': []}))

    def assertIP(self, expected, networks):

        class Nova(object):

            def __init__(self, networks):
                self.networks = networks

        nova = Nova(dict(a_name=networks))
        self.assertEqual(expected, self.bsj.get_nova_ip(nova))

    def test_nova_ip_malformed(self):
        self.assertIP(None, None)
        self.assertIP(None, [])

    def test_nova_ip_well_formed(self):
        self.assertIP('15.125.112.72', ['192.168.0.1', '15.125.112.72'])


class TestPrivateIP(unittest.TestCase):

    def setUp(self):
        super(TestPrivateIP, self).setUp()
        self.bsj = features.BootstrappedJuju()

    def assertPrivate(self, ip):
        self.assertTrue(self.bsj.is_ip_private(ip))

    def assertNotPrivate(self, ip):
        self.assertFalse(self.bsj.is_ip_private(ip))

    def test_public(self):
        self.assertNotPrivate('2.2.2.2')

    def test_class_A_private(self):
        self.assertPrivate('10.0.4.5')

    def test_class_B_public(self):
        self.assertNotPrivate('172.15.2.245')

    def test_class_B_private(self):
        self.assertPrivate('172.24.12.36')

    def test_class_C_private(self):
        self.assertPrivate('192.168.0.16')


@features.requires(features.swift)
class TestJujuBucket(unittest.TestCase):

    def test_it(self):
        pass
